@@ -4,6 +4,7 @@ import asyncio
 import contextlib
 import logging
 import struct
+import time
 from collections.abc import Awaitable, Callable, Iterable
 from dataclasses import replace
 from typing import Any, TypeVar, cast
@@ -45,10 +46,18 @@ _ADV_LOGGER = logging.getLogger("yalexs_ble_adv")
 
 WrapFuncType = TypeVar("WrapFuncType", bound=Callable[..., Any])
 
+NEVER_TIME = -86400.0
+
 DEFAULT_ATTEMPTS = 4
 
 DISCONNECT_DELAY = 5.1
 
+RESYNC_DELAY = 0.01
+
+# After a lock operation we need to wait for the lock to
+# update its state or it will return a stale state.
+LOCK_STALE_STATE_DEBOUNCE_DELAY = 4.0
+
 # How long to wait before processing an advertisement change
 ADV_UPDATE_COALESCE_SECONDS = 0.05
 
@@ -75,6 +84,8 @@ RETRY_EXCEPTIONS = (ResponseError, *BLEAK_RETRY_EXCEPTIONS)
 # there is no update from the lock.
 VALID_ADV_VALUES = {0, 1}
 
+AUTH_FAILURE_TO_START_REAUTH = 5
+
 
 def operation_lock(func: WrapFuncType) -> WrapFuncType:
     """Define a wrapper to only allow a single operation at a time."""
@@ -107,7 +118,13 @@ def retry_bluetooth_connection_error(func: WrapFuncType) -> WrapFuncType:
             try:
                 return await func(self, *args, **kwargs)
             except AuthError:
-                self._update_any_state([AuthState(successful=False)])
+                self._auth_failures += 1
+                if self._auth_failures >= AUTH_FAILURE_TO_START_REAUTH:
+                    # If the bluetooth connection drops in the middle of authentication
+                    # we may see it as a failed authentication. If we see 5 failed
+                    # authentications in a row we can reasonably assume that the key has
+                    # changed and we should re-authenticate.
+                    self._update_any_state([AuthState(successful=False)])
                 raise
             except BleakNotFoundError:
                 # The lock cannot be found so there is no
@@ -201,16 +218,17 @@ class PushLock:
         self._callbacks: list[
             Callable[[LockState, LockInfo, ConnectionInfo], None]
         ] = []
-        self._debounce_lock = asyncio.Lock()
+        self._update_task: asyncio.Task[None] | None = None
         self.loop = asyncio._get_running_loop()
         self._cancel_deferred_update: asyncio.TimerHandle | None = None
-        self._cancel_post_lock_op_sync: asyncio.TimerHandle | None = None
         self._client: Lock | None = None
         self._connect_lock = asyncio.Lock()
         self._disconnect_timer: asyncio.TimerHandle | None = None
         self._idle_disconnect_delay = idle_disconnect_delay
         self._first_update_future: asyncio.Future[None] | None = None
         self._background_tasks: set[asyncio.Task[None]] = set()
+        self._auth_failures: int = 0
+        self._last_lock_operation_complete_time = NEVER_TIME
 
     @property
     def local_name(self) -> str | None:
@@ -336,13 +354,27 @@ class PushLock:
             "%s: Executing forced disconnect",
             self.name,
         )
+        if (update_task := self._update_task) and not update_task.done():
+            self._update_task = None
+            update_task.cancel()
+            with contextlib.suppress(Exception, asyncio.CancelledError):
+                await update_task
         await self._execute_disconnect()
 
     def _disconnect_with_timer(self) -> None:
         """Disconnect from device."""
         if self._operation_lock.locked():
+            _LOGGER.debug("%s: Disconnect timer reset due to operation lock", self.name)
             self._reset_disconnect_timer()
             return
+        if self._cancel_deferred_update:
+            _LOGGER.debug(
+                "%s: Disconnect timer fired while we were waiting to update", self.name
+            )
+            self._reset_disconnect_timer()
+            self._cancel_future_update()
+            self._deferred_update()
+            return
         self._cancel_disconnect_timer()
         self.background_task(self._execute_timed_disconnect())
 
@@ -392,7 +424,10 @@ class PushLock:
             self._client = self._get_lock_instance()
             try:
                 await self._client.connect()
-            except Exception:
+            except Exception as ex:
+                _LOGGER.debug(
+                    "%s: Failed to connect due to %s, forcing disconnect", self.name, ex
+                )
                 await self._client.disconnect()
                 raise
             self._reset_disconnect_timer()
@@ -401,48 +436,39 @@ class PushLock:
     async def lock(self) -> None:
         """Lock the lock."""
         self._update_any_state([LockStatus.LOCKING])
-        self._cancel_update()
-        await self._lock_with_op_lock()
-
-    @operation_lock
-    @retry_bluetooth_connection_error
-    async def _lock_with_op_lock(self) -> None:
-        """Lock the lock."""
-        _LOGGER.debug("%s: Starting lock", self.name)
-        self._update_any_state([LockStatus.LOCKING])
-        self._cancel_update()
-        try:
-            lock = await self._ensure_connected()
-            self._cancel_update()
-            await lock.force_lock()
-        except Exception:
-            self._update_any_state([LockStatus.UNKNOWN])
-            raise
-        self._update_any_state([LockStatus.LOCKED])
-        _LOGGER.debug("%s: Finished lock", self.name)
+        self._cancel_future_update()
+        await self._execute_lock_operation(
+            "force_lock", LockStatus.LOCKING, LockStatus.LOCKED
+        )
 
     async def unlock(self) -> None:
         """Unlock the lock."""
         self._update_any_state([LockStatus.UNLOCKING])
-        self._cancel_update()
-        await self._unlock_with_op_lock()
+        self._cancel_future_update()
+        await self._execute_lock_operation(
+            "force_unlock", LockStatus.UNLOCKING, LockStatus.UNLOCKED
+        )
 
     @operation_lock
     @retry_bluetooth_connection_error
-    async def _unlock_with_op_lock(self) -> None:
-        """Unlock the lock."""
-        _LOGGER.debug("%s: Starting unlock", self.name)
-        self._update_any_state([LockStatus.UNLOCKING])
-        self._cancel_update()
+    async def _execute_lock_operation(
+        self, op_attr: str, pending_state: LockStatus, complete_state: LockStatus
+    ) -> None:
+        """Execute a lock operation."""
+        _LOGGER.debug("%s: Starting %s", self.name, pending_state)
+        self._update_any_state([pending_state])
+        self._cancel_future_update()
         try:
             lock = await self._ensure_connected()
-            self._cancel_update()
-            await lock.force_unlock()
+            self._cancel_future_update()
+            await getattr(lock, op_attr)()
         except Exception:
             self._update_any_state([LockStatus.UNKNOWN])
             raise
-        self._update_any_state([LockStatus.UNLOCKED])
-        _LOGGER.debug("%s: Finished unlock", self.name)
+        self._update_any_state([complete_state])
+        _LOGGER.debug("%s: Finished %s", self.name, complete_state)
+        self._last_lock_operation_complete_time = time.monotonic()
+        self._reset_disconnect_timer()
 
     def _state_callback(
         self, states: Iterable[LockStatus | DoorStatus | BatteryState]
@@ -483,13 +509,13 @@ class PushLock:
             and (not lock_state.auth or lock_state.auth.successful)
             and original_lock_status != LockStatus.UNKNOWN
         ):
-            self._schedule_resync()
+            self._schedule_future_update(RESYNC_DELAY)
 
         self._callback_state(lock_state)
 
     async def update(self) -> None:
         """Request that status be updated."""
-        self._schedule_update(
+        self._schedule_future_update_with_debounce(
             0
             if self._client and self._client.is_connected
             else MANUAL_UPDATE_COALESCE_SECONDS
@@ -516,9 +542,6 @@ class PushLock:
                 self._lock_info = await lock.lock_info()
             state = await lock.status()
             battery_state = await lock.battery()
-            state = replace(
-                state, battery=battery_state, auth=AuthState(successful=True)
-            )
         except asyncio.CancelledError:
             _LOGGER.debug(
                 "%s: In-progress update canceled due "
@@ -526,6 +549,11 @@ class PushLock:
                 self.name,
             )
             raise
+        else:
+            self._auth_failures = 0
+            state = replace(
+                state, battery=battery_state, auth=AuthState(successful=True)
+            )
         _LOGGER.debug("%s: Finished update", self.name)
         self._callback_state(state)
 
@@ -634,7 +662,7 @@ class PushLock:
                 scheduled_update,
             )
         if next_update:
-            self._schedule_update(next_update)
+            self._schedule_future_update_with_debounce(next_update)
 
     async def start(self) -> Callable[[], None]:
         """Start watching for updates."""
@@ -645,11 +673,11 @@ class PushLock:
         self._first_update_future = asyncio.get_running_loop().create_future()
         if device := await get_device(self.address):
             self.set_ble_device(device)
-            self._schedule_update(ADV_UPDATE_COALESCE_SECONDS)
+            self._schedule_future_update_with_debounce(ADV_UPDATE_COALESCE_SECONDS)
 
         def _cancel() -> None:
             self._running = False
-            self._cancel_resync()
+            self._cancel_future_update()
             self.background_task(self._execute_forced_disconnect())
 
         return _cancel
@@ -677,31 +705,14 @@ class PushLock:
         finally:
             self._first_update_future = None
 
-    def _cancel_resync(self) -> None:
-        """Cancel a resync."""
-        if self._cancel_post_lock_op_sync:
-            self._cancel_post_lock_op_sync.cancel()
-            self._cancel_post_lock_op_sync = None
-
-    def _cancel_update(self) -> None:
+    def _cancel_future_update(self) -> None:
         """Cancel an update."""
         if self._cancel_deferred_update:
             self._cancel_deferred_update.cancel()
             self._cancel_deferred_update = None
 
-    def _resync(self) -> None:
-        """Resync the lock state."""
-        self._schedule_update(0.01)
-
-    def _schedule_resync(self) -> None:
-        """Schedule a resync."""
-        self._cancel_resync()
-        self._cancel_post_lock_op_sync = self.loop.call_later(
-            POST_OPERATION_SYNC_TIME, self._resync
-        )
-
-    def _schedule_update(self, seconds: float) -> None:
-        """Schedule an update."""
+    def _schedule_future_update_with_debounce(self, seconds: float) -> None:
+        """Schedule an update with a potential debounce."""
         now = self.loop.time()
         future_update_time = seconds
         if self._cancel_deferred_update:
@@ -723,26 +734,37 @@ class PushLock:
                 )
                 return
             _LOGGER.debug("%s: Rescheduling update", self.name)
-            self._cancel_update()
+        self._schedule_future_update(future_update_time)
+
+    def _schedule_future_update(self, future_update_time: float) -> None:
+        """Schedule an update in future seconds."""
         _LOGGER.debug(
             "%s: Scheduling update to happen in %s seconds",
             self.name,
             future_update_time,
         )
-        self._cancel_deferred_update = self.loop.call_at(
-            now + future_update_time, self._deferred_update
+        self._cancel_future_update()
+        self._cancel_deferred_update = self.loop.call_later(
+            future_update_time, self._deferred_update
         )
 
     def _deferred_update(self) -> None:
         """Update the lock state."""
-        self._cancel_deferred_update = None
-        if self._debounce_lock.locked():
+        self._cancel_future_update()
+        now = time.monotonic()
+        if self._update_task and not self._update_task.done():
             _LOGGER.debug(
                 "%s: Rescheduling update since one already in progress", self.name
             )
-            self._schedule_update(UPDATE_IN_PROGRESS_DEFER_SECONDS)
+            self._schedule_future_update_with_debounce(UPDATE_IN_PROGRESS_DEFER_SECONDS)
             return
-        self.loop.create_task(self._queue_update())
+        if (
+            seconds_time_lock_op := (now - self._last_lock_operation_complete_time)
+        ) < LOCK_STALE_STATE_DEBOUNCE_DELAY:
+            _LOGGER.debug("%s: Rescheduling update to avoid stale state", self.name)
+            self._schedule_future_update_with_debounce(seconds_time_lock_op)
+            return
+        self._update_task = asyncio.create_task(self._execute_deferred_update())
 
     def _set_update_state(self, exception: Exception | None) -> None:
         """Set the update state."""
@@ -753,44 +775,40 @@ class PushLock:
         else:
             self._first_update_future.set_result(None)
 
-    async def _queue_update(self) -> None:
-        """Watch for updates."""
-        _LOGGER.debug("%s: Update queued", self.name)
-        async with self._debounce_lock:
-            _LOGGER.debug("%s: Queued update starting", self.name)
-            if not self._running:
-                _LOGGER.debug(
-                    "%s: Queued updated ignored because not running", self.name
-                )
-                return
-            _LOGGER.debug("%s: Starting update", self.name)
-            try:
-                await self._update()
-                self._set_update_state(None)
-            except AuthError as ex:
-                self._set_update_state(ex)
-                _LOGGER.error(
-                    "%s: Auth error: key or slot (key index) is incorrect: %s",
-                    self.name,
-                    ex,
-                    exc_info=True,
-                )
-            except asyncio.CancelledError:
-                self._set_update_state(RuntimeError("Update was canceled"))
-                _LOGGER.debug("%s: In-progress update canceled", self.name)
-            except asyncio.TimeoutError as ex:
-                self._set_update_state(ex)
-                _LOGGER.exception("%s: Timed out updating", self.name)
-            except BleakError as ex:
-                wrapped_bleak_exc = BluetoothError(str(ex))
-                wrapped_bleak_exc.__cause__ = ex
-                self._set_update_state(wrapped_bleak_exc)
-                _LOGGER.exception("%s: Bluetooth error updating", self.name)
-            except Exception as ex:  # pylint: disable=broad-except
-                wrapped_exc = YaleXSBLEError(str(ex))
-                wrapped_exc.__cause__ = ex
-                self._set_update_state(wrapped_exc)
-                _LOGGER.exception("%s: Unknown error updating", self.name)
+    async def _execute_deferred_update(self) -> None:
+        """Execute deferred update."""
+        _LOGGER.debug("%s: Deferred update starting", self.name)
+        if not self._running:
+            _LOGGER.debug("%s: Deferred updated ignored because not running", self.name)
+            return
+        _LOGGER.debug("%s: Starting deferred update", self.name)
+        try:
+            await self._update()
+            self._set_update_state(None)
+        except AuthError as ex:
+            self._set_update_state(ex)
+            _LOGGER.error(
+                "%s: Auth error: key or slot (key index) is incorrect: %s",
+                self.name,
+                ex,
+                exc_info=True,
+            )
+        except asyncio.CancelledError:
+            self._set_update_state(RuntimeError("Update was canceled"))
+            _LOGGER.debug("%s: In-progress update canceled", self.name)
+        except asyncio.TimeoutError as ex:
+            self._set_update_state(ex)
+            _LOGGER.exception("%s: Timed out updating", self.name)
+        except BleakError as ex:
+            wrapped_bleak_exc = BluetoothError(str(ex))
+            wrapped_bleak_exc.__cause__ = ex
+            self._set_update_state(wrapped_bleak_exc)
+            _LOGGER.exception("%s: Bluetooth error updating", self.name)
+        except Exception as ex:  # pylint: disable=broad-except
+            wrapped_exc = YaleXSBLEError(str(ex))
+            wrapped_exc.__cause__ = ex
+            self._set_update_state(wrapped_exc)
+            _LOGGER.exception("%s: Unknown error updating", self.name)
 
 
 def get_homekit_state_num(data: bytes) -> int:
